@@ -3,7 +3,7 @@ use crate::{
     ExParseError,
 };
 use smallvec::{smallvec, SmallVec};
-use std::fmt::Debug;
+use std::{fmt, iter::repeat};
 
 type ExprIdxVec = SmallVec<[usize; 32]>;
 
@@ -17,19 +17,19 @@ pub type FlatOpVec<T> = SmallVec<[FlatOp<T>; N_NODES_ON_STACK]>;
 
 /// A `FlatOp` contains besides a binary operation an optional unary operation that
 /// will be executed after the binary operation in case of its existence.
-#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
 pub struct FlatOp<T: Copy> {
     unary_op: UnaryOp<T>,
     bin_op: BinOp<T>,
 }
 
-#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
 pub enum FlatNodeKind<T: Copy> {
     Num(T),
     Var(usize),
 }
 
-#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
 pub struct FlatNode<T: Copy> {
     kind: FlatNodeKind<T>,
     unary_op: UnaryOp<T>,
@@ -44,6 +44,124 @@ impl<T: Copy> FlatNode<T> {
     }
 }
 
+fn flatten_vecs<T: Copy>(
+    deep_expr: &DeepEx<T>,
+    prio_offset: i32,
+) -> (FlatNodeVec<T>, FlatOpVec<T>) {
+    let mut flat_nodes = FlatNodeVec::<T>::new();
+    let mut flat_ops = FlatOpVec::<T>::new();
+
+    for (node_idx, node) in deep_expr.nodes.iter().enumerate() {
+        match node {
+            DeepNode::Num(num) => {
+                let flat_node = FlatNode::from_kind(FlatNodeKind::Num(*num));
+                flat_nodes.push(flat_node);
+            }
+            DeepNode::Var(idx) => {
+                let flat_node = FlatNode::from_kind(FlatNodeKind::Var(*idx));
+                flat_nodes.push(flat_node);
+            }
+            DeepNode::Expr(e) => {
+                let (mut sub_nodes, mut sub_ops) = flatten_vecs(e, prio_offset + 100i32);
+                flat_nodes.append(&mut sub_nodes);
+                flat_ops.append(&mut sub_ops);
+            }
+        };
+        if node_idx < deep_expr.bin_ops.ops.len() {
+            let prio_adapted_bin_op = BinOp {
+                apply: deep_expr.bin_ops.ops[node_idx].apply,
+                prio: deep_expr.bin_ops.ops[node_idx].prio + prio_offset,
+            };
+            flat_ops.push(FlatOp {
+                bin_op: prio_adapted_bin_op,
+                unary_op: UnaryOp::new(),
+            });
+        }
+    }
+
+    if deep_expr.unary_op.op.len() > 0 {
+        if flat_ops.len() > 0 {
+            // find the last binary operator with the lowest priority of this expression,
+            // since this will be executed as the last one
+            let low_prio_op = match flat_ops.iter_mut().rev().min_by_key(|op| op.bin_op.prio) {
+                None => panic!("cannot have more than one flat node but no binary ops"),
+                Some(x) => x,
+            };
+            low_prio_op
+                .unary_op
+                .append_front(&mut deep_expr.unary_op.op.clone());
+        } else {
+            flat_nodes[0]
+                .unary_op
+                .append_front(&mut deep_expr.unary_op.op.clone());
+        }
+    }
+    (flat_nodes, flat_ops)
+}
+
+pub fn flatten<T: Copy>(deep_ex: DeepEx<T>) -> FlatEx<T> {
+    let (nodes, ops) = flatten_vecs(&deep_ex, 0);
+    let indices = prioritized_indices_flat(&ops, &nodes);
+    let mut found_vars = SmallVec::<[usize; 16]>::new();
+    let n_unique_vars = nodes
+        .iter()
+        .filter_map(|n| match n.kind {
+            FlatNodeKind::Var(idx) => {
+                if !found_vars.contains(&idx) {
+                    found_vars.push(idx);
+                    Some(idx)
+                } else {
+                    None
+                }
+            }
+            _ => None,
+        })
+        .count();
+    FlatEx {
+        nodes: nodes,
+        ops: ops,
+        prio_indices: indices,
+        n_unique_vars: n_unique_vars,
+        deepex: Some(deep_ex),
+    }
+}
+
+fn prioritized_indices_flat<T: Copy>(ops: &[FlatOp<T>], nodes: &FlatNodeVec<T>) -> ExprIdxVec {
+    let prio_increase =
+        |bin_op_idx: usize| match (&nodes[bin_op_idx].kind, &nodes[bin_op_idx + 1].kind) {
+            (FlatNodeKind::Num(_), FlatNodeKind::Num(_)) => {
+                let prio_inc = 5;
+                &ops[bin_op_idx].bin_op.prio * 10 + prio_inc
+            }
+            _ => &ops[bin_op_idx].bin_op.prio * 10,
+        };
+    let mut indices: ExprIdxVec = (0..ops.len()).collect();
+    indices.sort_by(|i1, i2| {
+        let prio_i1 = prio_increase(*i1);
+        let prio_i2 = prio_increase(*i2);
+        prio_i2.partial_cmp(&prio_i1).unwrap()
+    });
+    indices
+}
+
+fn prioritized_indices<T: Copy>(bin_ops: &[BinOp<T>], nodes: &[DeepNode<T>]) -> ExprIdxVec {
+    let prio_increase = |bin_op_idx: usize| match (&nodes[bin_op_idx], &nodes[bin_op_idx + 1]) {
+        (DeepNode::Num(_), DeepNode::Num(_)) => {
+            let prio_inc = 5;
+            &bin_ops[bin_op_idx].prio * 10 + prio_inc
+        }
+        _ => &bin_ops[bin_op_idx].prio * 10,
+    };
+
+    let mut indices: ExprIdxVec = (0..bin_ops.len()).collect();
+    indices.sort_by(|i1, i2| {
+        let prio_i1 = prio_increase(*i1);
+        let prio_i2 = prio_increase(*i2);
+        prio_i2.partial_cmp(&prio_i1).unwrap()
+    });
+    indices
+}
+
 /// This is the core data type representing a flattened expression and the result of
 /// parsing a string. We use flattened expressions to make efficient evaluation possible.
 /// Simplified, a flat expression consists of a [`SmallVec`](SmallVec) of nodes and a
@@ -73,14 +191,15 @@ impl<T: Copy> FlatNode<T> {
 /// operators, and no parentheses.
 ///
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
-pub struct FlatEx<T: Copy> {
+pub struct FlatEx<'a, T: Copy> {
     nodes: FlatNodeVec<T>,
     ops: FlatOpVec<T>,
     prio_indices: ExprIdxVec,
     n_unique_vars: usize,
+    deepex: Option<DeepEx<'a, T>>,
 }
 
-impl<T: Copy + Debug> FlatEx<T> {
+impl<'a, T: Copy + fmt::Debug> FlatEx<'a, T> {
     /// Evaluates an expression with the given variable values and returns the computed
     /// result.
     ///
@@ -136,125 +255,88 @@ impl<T: Copy + Debug> FlatEx<T> {
         }
         Ok(numbers[0])
     }
+
+    /// Creates an expression string that corresponds to the `FlatEx` instance. This is
+    /// not necessarily the input string. More precisely,
+    /// * variable names are forgotten,
+    /// * variables are put into curly braces, and
+    /// * expressions will be put between parentheses, e.g.,
+    /// ```rust
+    /// # use std::error::Error;
+    /// # fn main() -> Result<(), Box<dyn Error>> {
+    /// #
+    /// use exmex::parse_with_default_ops;
+    /// let flatex = parse_with_default_ops::<f64>("--sin(z)")?;
+    /// assert_eq!(format!("{}", flatex), "-(-(sin({x0})))");
+    /// #
+    /// #     Ok(())
+    /// # }
+    /// ```
+    /// 
+    pub fn unparse(&self) -> Result<String, ExParseError> {
+        match &self.deepex {
+            Some(deepex) => Ok(deepex.unparse()),
+            None => Err(ExParseError {
+                msg: "unparse impossible, since deep expression optimized away".to_string(),
+            }),
+        }
+    }
+    /// Usually, a `FlatEx` instance keeps a nested, deep structure of the expression. This functions removes
+    /// the deep expression to reduce memory consumption. [`unparse`](FlatEx::unparse) is not
+    /// possible anymore afterwards.
+    pub fn clear_deepex(&mut self) {
+        self.deepex = None;
+    }
 }
 
-fn flatten_vecs<T: Copy>(
-    deep_expr: &DeepEx<T>,
-    prio_offset: i32,
-) -> (FlatNodeVec<T>, FlatOpVec<T>) {
-    let mut flat_nodes = FlatNodeVec::<T>::new();
-    let mut flat_ops = FlatOpVec::<T>::new();
-
-    for (node_idx, node) in deep_expr.nodes.iter().enumerate() {
-        match node {
-            DeepNode::Num(num) => {
-                let flat_node = FlatNode::from_kind(FlatNodeKind::Num(*num));
-                flat_nodes.push(flat_node);
-            }
-            DeepNode::Var(idx) => {
-                let flat_node = FlatNode::from_kind(FlatNodeKind::Var(*idx));
-                flat_nodes.push(flat_node);
-            }
-            DeepNode::Expr(e) => {
-                let (mut sub_nodes, mut sub_ops) = flatten_vecs(e, prio_offset + 100i32);
-                flat_nodes.append(&mut sub_nodes);
-                flat_ops.append(&mut sub_ops);
-            }
-        };
-        if node_idx < deep_expr.bin_ops.len() {
-            let prio_adapted_bin_op = BinOp {
-                apply: deep_expr.bin_ops[node_idx].apply,
-                prio: deep_expr.bin_ops[node_idx].prio + prio_offset,
-            };
-            flat_ops.push(FlatOp {
-                bin_op: prio_adapted_bin_op,
-                unary_op: UnaryOp::new(),
-            });
+/// The expression is displayed as a string created by [`unparse`](FlatEx::unparse).
+impl<'a, T: Copy + fmt::Debug> fmt::Display for FlatEx<'a, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match &self.deepex {
+            None => write!(f, "[FlatEx display information not available]"),
+            Some(deepex) => write!(f, "{}", deepex.unparse()),
         }
     }
-
-    if deep_expr.unary_op.len() > 0 {
-        if flat_ops.len() > 0 {
-            // find the last binary operator with the lowest priority of this expression,
-            // since this will be executed as the last one
-            let low_prio_op = match flat_ops.iter_mut().rev().min_by_key(|op| op.bin_op.prio) {
-                None => panic!("cannot have more than one flat node but no binary ops"),
-                Some(x) => x,
-            };
-            low_prio_op
-                .unary_op
-                .append_front(&mut deep_expr.unary_op.clone());
-        } else {
-            flat_nodes[0]
-                .unary_op
-                .append_front(&mut deep_expr.unary_op.clone());
-        }
-    }
-    (flat_nodes, flat_ops)
 }
 
 /// A deep node can be an expression, a number, or
 /// a variable.
-#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
-pub enum DeepNode<T: Copy> {
-    Expr(DeepEx<T>),
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
+pub enum DeepNode<'a, T: Copy> {
+    Expr(DeepEx<'a, T>),
     Num(T),
     /// The contained integer points to the index of the variable in the slice of
     /// variables passed to [`eval`](Expression::eval).
     Var(usize),
 }
 
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
+pub struct BinOpsWithReprs<'a, T: Copy> {
+    pub reprs: Vec<&'a str>,
+    pub ops: BinOpVec<T>,
+}
+
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
+pub struct UnaryOpWithReprs<'a, T: Copy> {
+    pub reprs: Vec<&'a str>,
+    pub op: UnaryOp<T>,
+}
+
 /// A deep expression evaluates co-recursively since its nodes can contain other deep
 /// expressions.
-#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
-pub struct DeepEx<T: Copy> {
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, fmt::Debug)]
+pub struct DeepEx<'a, T: Copy> {
     /// Nodes can be numbers, variables, or other expressions.
-    nodes: Vec<DeepNode<T>>,
+    nodes: Vec<DeepNode<'a, T>>,
     /// Binary operators applied to the nodes according to their priority.
-    bin_ops: BinOpVec<T>,
+    bin_ops: BinOpsWithReprs<'a, T>,
     /// Unary operators are applied to the result of evaluating all nodes with all
     /// binary operators.
-    unary_op: UnaryOp<T>,
+    unary_op: UnaryOpWithReprs<'a, T>,
     prio_indices: ExprIdxVec,
 }
 
-fn prioritized_indices_flat<T: Copy>(ops: &[FlatOp<T>], nodes: &FlatNodeVec<T>) -> ExprIdxVec {
-    let prio_increase =
-        |bin_op_idx: usize| match (&nodes[bin_op_idx].kind, &nodes[bin_op_idx + 1].kind) {
-            (FlatNodeKind::Num(_), FlatNodeKind::Num(_)) => {
-                let prio_inc = 5;
-                &ops[bin_op_idx].bin_op.prio * 10 + prio_inc
-            }
-            _ => &ops[bin_op_idx].bin_op.prio * 10,
-        };
-    let mut indices: ExprIdxVec = (0..ops.len()).collect();
-    indices.sort_by(|i1, i2| {
-        let prio_i1 = prio_increase(*i1);
-        let prio_i2 = prio_increase(*i2);
-        prio_i2.partial_cmp(&prio_i1).unwrap()
-    });
-    indices
-}
-
-fn prioritized_indices<T: Copy>(bin_ops: &[BinOp<T>], nodes: &[DeepNode<T>]) -> ExprIdxVec {
-    let prio_increase = |bin_op_idx: usize| match (&nodes[bin_op_idx], &nodes[bin_op_idx + 1]) {
-        (DeepNode::Num(_), DeepNode::Num(_)) => {
-            let prio_inc = 5;
-            &bin_ops[bin_op_idx].prio * 10 + prio_inc
-        }
-        _ => &bin_ops[bin_op_idx].prio * 10,
-    };
-
-    let mut indices: ExprIdxVec = (0..bin_ops.len()).collect();
-    indices.sort_by(|i1, i2| {
-        let prio_i1 = prio_increase(*i1);
-        let prio_i2 = prio_increase(*i2);
-        prio_i2.partial_cmp(&prio_i1).unwrap()
-    });
-    indices
-}
-
-impl<T: Copy + Debug> DeepEx<T> {
+impl<'a, T: Copy + fmt::Debug> DeepEx<'a, T> {
     /// Evaluates all operators with numbers as operands.
     pub fn compile(&mut self) {
         // change from exression to number if an expression contains only a number
@@ -271,7 +353,7 @@ impl<T: Copy + Debug> DeepEx<T> {
             };
         }
         // after changing from expressions to numbers where possible the prios might change
-        self.prio_indices = prioritized_indices(&self.bin_ops, &self.nodes);
+        self.prio_indices = prioritized_indices(&self.bin_ops.ops, &self.nodes);
 
         let mut num_inds = self.prio_indices.clone();
         let mut used_prio_indices = ExprIdxVec::new();
@@ -280,7 +362,7 @@ impl<T: Copy + Debug> DeepEx<T> {
             let node_1 = &self.nodes[num_idx];
             let node_2 = &self.nodes[num_idx + 1];
             if let (DeepNode::Num(num_1), DeepNode::Num(num_2)) = (node_1, node_2) {
-                let bin_op_result = (self.bin_ops[bin_op_idx].apply)(*num_1, *num_2);
+                let bin_op_result = (self.bin_ops.ops[bin_op_idx].apply)(*num_1, *num_2);
                 self.nodes[num_idx] = DeepNode::Num(bin_op_result);
                 self.nodes.remove(num_idx + 1);
                 // reduce indices after removed position
@@ -295,8 +377,9 @@ impl<T: Copy + Debug> DeepEx<T> {
             }
         }
 
-        self.bin_ops = self
+        self.bin_ops.ops = self
             .bin_ops
+            .ops
             .iter()
             .enumerate()
             .filter(|(i, _)| !used_prio_indices.contains(i))
@@ -306,25 +389,26 @@ impl<T: Copy + Debug> DeepEx<T> {
         if self.nodes.len() == 1 {
             match self.nodes[0] {
                 DeepNode::Num(n) => {
-                    self.nodes[0] = DeepNode::Num(self.unary_op.apply(n));
-                    self.unary_op.clear();
+                    self.nodes[0] = DeepNode::Num(self.unary_op.op.apply(n));
+                    self.unary_op.op.clear();
+                    self.unary_op.reprs.clear();
                 }
                 _ => (),
             }
         }
-        self.prio_indices = prioritized_indices(&self.bin_ops, &self.nodes);
+        self.prio_indices = prioritized_indices(&self.bin_ops.ops, &self.nodes);
     }
     pub fn new(
-        nodes: Vec<DeepNode<T>>,
-        bin_ops: BinOpVec<T>,
-        unary_op: UnaryOp<T>,
-    ) -> Result<DeepEx<T>, ExParseError> {
-        if nodes.len() != bin_ops.len() + 1 {
+        nodes: Vec<DeepNode<'a, T>>,
+        bin_ops: BinOpsWithReprs<'a, T>,
+        unary_op: UnaryOpWithReprs<'a, T>,
+    ) -> Result<DeepEx<'a, T>, ExParseError> {
+        if nodes.len() != bin_ops.ops.len() + 1 {
             Err(ExParseError {
                 msg: "mismatch between number of nodes and binary operators".to_string(),
             })
         } else {
-            let indices = prioritized_indices(&bin_ops, &nodes);
+            let indices = prioritized_indices(&bin_ops.ops, &nodes);
             let mut expr = DeepEx {
                 nodes: nodes,
                 bin_ops: bin_ops,
@@ -335,68 +419,105 @@ impl<T: Copy + Debug> DeepEx<T> {
             Ok(expr)
         }
     }
-    pub fn flatten(&self) -> FlatEx<T> {
-        let (nodes, ops) = flatten_vecs(self, 0);
-        let indices = prioritized_indices_flat(&ops, &nodes);
-        let mut found_vars = SmallVec::<[usize; 16]>::new();
-        let n_unique_vars = nodes
-            .iter()
-            .filter_map(|n| match n.kind {
-                FlatNodeKind::Var(idx) => {
-                    if !found_vars.contains(&idx) {
-                        found_vars.push(idx);
-                        Some(idx)
-                    } else {
-                        None
-                    }
+
+    pub fn unparse(&self) -> String {
+        let mut node_strings = self.nodes.iter().map(|n| match n {
+            DeepNode::Num(n) => format!("{:?}", n),
+            DeepNode::Var(idx) => format!("{{x{}}}", idx),
+            DeepNode::Expr(e) => {
+                if e.unary_op.op.len() == 0 {
+                    format!("({})", e.unparse())
+                } else {
+                    e.unparse()
                 }
-                _ => None,
-            })
-            .count();
-        FlatEx {
-            nodes: nodes,
-            ops: ops,
-            prio_indices: indices,
-            n_unique_vars: n_unique_vars,
+            }
+        });
+        let mut bin_op_strings = self.bin_ops.reprs.iter();
+        // a valid expression has at least one node
+        let first_node_str = node_strings.next().unwrap();
+        let node_with_bin_ops_string = node_strings.fold(first_node_str, |mut res, node_str| {
+            let bin_op_str = bin_op_strings.next().unwrap();
+            res.push_str(bin_op_str);
+            res.push_str(node_str.as_str());
+            res
+        });
+        let unary_op_string = self
+            .unary_op
+            .reprs
+            .iter()
+            .fold(String::new(), |mut res, uop_str| {
+                res.push_str(uop_str);
+                res.push_str("(");
+                res
+            });
+        let closings =
+            repeat(")")
+                .take(self.unary_op.op.len())
+                .fold(String::new(), |mut res, closing| {
+                    res.push_str(closing);
+                    res
+                });
+        if self.unary_op.op.len() == 0 {
+            node_with_bin_ops_string
+        } else {
+            format!(
+                "{}{}{}",
+                unary_op_string, node_with_bin_ops_string, closings
+            )
         }
     }
 }
 
+impl<'a, T: Copy + fmt::Debug> fmt::Display for DeepEx<'a, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "{}", self.unparse())
+    }
+}
+
 #[cfg(test)]
 use crate::{make_default_operators, parse_with_default_ops, util::assert_float_eq_f64};
 #[test]
+fn test_flat_clear() {
+    let mut flatex = parse_with_default_ops::<f64>("x*(2*(2*(2*4*8)))").unwrap();
+    assert!(flatex.deepex.is_some());
+    flatex.clear_deepex();
+    assert!(flatex.deepex.is_none());
+    assert_float_eq_f64(flatex.eval(&[1.0]).unwrap(), 2.0 * 2.0 * 2.0 * 4.0 * 8.0);
+    assert_eq!(flatex.nodes.len(), 2);
+}
+#[test]
 fn test_flat_compile() {
-    let flat_ex = parse_with_default_ops::<f64>("1*sin(2-0.1)").unwrap();
-    assert_float_eq_f64(flat_ex.eval(&[]).unwrap(), 1.9f64.sin());
-    assert_eq!(flat_ex.nodes.len(), 1);
+    let flatex = parse_with_default_ops::<f64>("1*sin(2-0.1)").unwrap();
+    assert_float_eq_f64(flatex.eval(&[]).unwrap(), 1.9f64.sin());
+    assert_eq!(flatex.nodes.len(), 1);
 
-    let flat_ex = parse_with_default_ops::<f64>("x*(2*(2*(2*4*8)))").unwrap();
-    assert_float_eq_f64(flat_ex.eval(&[1.0]).unwrap(), 2.0 * 2.0 * 2.0 * 4.0 * 8.0);
-    assert_eq!(flat_ex.nodes.len(), 2);
+    let flatex = parse_with_default_ops::<f64>("x*(2*(2*(2*4*8)))").unwrap();
+    assert_float_eq_f64(flatex.eval(&[1.0]).unwrap(), 2.0 * 2.0 * 2.0 * 4.0 * 8.0);
+    assert_eq!(flatex.nodes.len(), 2);
 
-    let flat_ex = parse_with_default_ops::<f64>("1*sin(2-0.1) + x").unwrap();
-    assert_float_eq_f64(flat_ex.eval(&[0.0]).unwrap(), 1.9f64.sin());
-    assert_eq!(flat_ex.nodes.len(), 2);
-    match flat_ex.nodes[0].kind {
+    let flatex = parse_with_default_ops::<f64>("1*sin(2-0.1) + x").unwrap();
+    assert_float_eq_f64(flatex.eval(&[0.0]).unwrap(), 1.9f64.sin());
+    assert_eq!(flatex.nodes.len(), 2);
+    match flatex.nodes[0].kind {
         FlatNodeKind::Num(n) => assert_float_eq_f64(n, 1.9f64.sin()),
         _ => assert!(false),
     }
-    match flat_ex.nodes[1].kind {
+    match flatex.nodes[1].kind {
         FlatNodeKind::Var(idx) => assert_eq!(idx, 0),
         _ => assert!(false),
     }
 
-    let flat_ex = parse_with_default_ops::<f64>("y + 1 - cos(1/(1*sin(2-0.1))-2) + 2 + x").unwrap();
-    assert_eq!(flat_ex.nodes.len(), 3);
-    match flat_ex.nodes[0].kind {
+    let flatex = parse_with_default_ops::<f64>("y + 1 - cos(1/(1*sin(2-0.1))-2) + 2 + x").unwrap();
+    assert_eq!(flatex.nodes.len(), 3);
+    match flatex.nodes[0].kind {
         FlatNodeKind::Var(idx) => assert_eq!(idx, 0),
         _ => assert!(false),
     }
-    match flat_ex.nodes[1].kind {
+    match flatex.nodes[1].kind {
         FlatNodeKind::Num(_) => (),
         _ => assert!(false),
     }
-    match flat_ex.nodes[2].kind {
+    match flatex.nodes[2].kind {
         FlatNodeKind::Var(idx) => assert_eq!(idx, 1),
         _ => assert!(false),
     }
@@ -406,11 +527,24 @@ fn test_flat_compile() {
 fn test_deep_compile() {
     let ops = make_default_operators();
     let nodes = vec![DeepNode::Num(4.5), DeepNode::Num(0.5), DeepNode::Num(1.4)];
-    let bin_ops: BinOpVec<f64> = smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()];
-    let unary_op = UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]);
+    let bin_ops = BinOpsWithReprs {
+        reprs: vec![ops[1].repr, ops[3].repr],
+        ops: smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()],
+    };
+    let unary_op = UnaryOpWithReprs {
+        reprs: vec![ops[6].repr],
+        op: UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]),
+    };
     let deep_ex = DeepEx::new(nodes, bin_ops, unary_op).unwrap();
-    let bin_ops: BinOpVec<f64> = smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()];
-    let unary_op = UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]);
+
+    let bin_ops = BinOpsWithReprs {
+        reprs: vec![ops[1].repr, ops[3].repr],
+        ops: smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()],
+    };
+    let unary_op = UnaryOpWithReprs {
+        reprs: vec![ops[6].repr],
+        op: UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]),
+    };
     let nodes = vec![
         DeepNode::Num(4.5),
         DeepNode::Num(0.5),
@@ -419,7 +553,7 @@ fn test_deep_compile() {
     let deep_ex = DeepEx::new(nodes, bin_ops, unary_op).unwrap();
     assert_eq!(deep_ex.nodes.len(), 1);
     match deep_ex.nodes[0] {
-        DeepNode::Num(n) => assert_eq!(deep_ex.unary_op.apply(n), n),
+        DeepNode::Num(n) => assert_eq!(deep_ex.unary_op.op.apply(n), n),
         _ => {
             assert!(false);
         }
