@@ -1,4 +1,7 @@
-use crate::expression::{BinOpVec, DeepEx, DeepNode, FlatEx, N_NODES_ON_STACK};
+use crate::expression::{
+    flatten, BinOpVec, BinOpsWithReprs, DeepEx, DeepNode, FlatEx, UnaryOpWithReprs,
+    N_NODES_ON_STACK,
+};
 use crate::operators::{make_default_operators, BinOp, Operator, UnaryOp, VecOfUnaryFuncs};
 // use itertools::Itertools;
 use lazy_static::lazy_static;
@@ -76,7 +79,7 @@ fn is_numeric_regex<'a>(re: &Regex, text: &'a str) -> Option<&'a str> {
 ///
 /// See [`parse_with_number_pattern`](parse_with_number_pattern)
 ///
-fn parsed_tokens<'a, 'b, T: Copy + FromStr + Debug, F: Fn(&'b str) -> Option<&'b str>>(
+fn tokenize_and_analyze<'a, 'b, T: Copy + FromStr + Debug, F: Fn(&'b str) -> Option<&'b str>>(
     text: &'b str,
     ops_in: &[Operator<'a, T>],
     is_numeric: F,
@@ -126,7 +129,7 @@ where
             } else if c == '{' {
                 let n_count = text_rest.chars().take_while(|c| *c != '}').count();
                 cur_offset += n_count + 1;
-                ParsedToken::<T>::Var(text_rest[1..n_count + 1].to_string())
+                ParsedToken::<T>::Var(text_rest[1..n_count].to_string())
             } else if {
                 maybe_num = is_numeric(text_rest);
                 maybe_num.is_some()
@@ -174,11 +177,11 @@ where
 ///
 /// See [`parse_with_number_pattern`](parse_with_number_pattern)
 ///
-fn make_expression<T>(
-    parsed_tokens: &[ParsedToken<T>],
+fn make_expression<'a, T>(
+    parsed_tokens: &[ParsedToken<'a, T>],
     parsed_vars: &[&String],
-    unary_ops: UnaryOp<T>,
-) -> Result<(DeepEx<T>, usize), ExParseError>
+    unary_ops: UnaryOpWithReprs<'a, T>,
+) -> Result<(DeepEx<'a, T>, usize), ExParseError>
 where
     T: Copy + FromStr + Debug,
 {
@@ -206,19 +209,25 @@ where
     };
     // this closure handles the case that a token is a unary operator and accesses the
     // variable 'tokens' from the outer scope
-    let process_unary = |i: usize, uo| {
+    let process_unary = |i: usize, uo, repr| {
         // gather subsequent unary operators from the beginning
-        let vec_of_uops = once(uo)
-            .chain(
-                (i + 1..parsed_tokens.len())
-                    .map(|j| match parsed_tokens[j] {
-                        ParsedToken::Op(op) => op.unary_op,
-                        _ => None,
-                    })
-                    .take_while(|uo_| uo_.is_some())
-                    .flatten(),
-            )
+        let iter_of_uops = once((repr, uo)).chain(
+            (i + 1..parsed_tokens.len())
+                .map(|j| match parsed_tokens[j] {
+                    ParsedToken::Op(op) => (op.repr, op.unary_op),
+                    _ => ("", None),
+                })
+                .take_while(|(_, uo_)| uo_.is_some())
+                .map(|(repr_, uo_)| (repr_, uo_.unwrap())),
+        );
+        let vec_of_uops = iter_of_uops
+            .clone()
+            .map(|(_, uo_)| uo_)
             .collect::<VecOfUnaryFuncs<_>>();
+        let vec_of_uop_reprs = iter_of_uops
+            .clone()
+            .map(|(repr_, _)| repr_)
+            .collect::<Vec<_>>();
         let n_uops = vec_of_uops.len();
         let uop = UnaryOp::from_vec(vec_of_uops);
         match &parsed_tokens[i + n_uops] {
@@ -227,16 +236,28 @@ where
                     msg: "closing parenthesis after an operator".to_string(),
                 }),
                 Paren::Open => {
-                    let (expr, i_forward) =
-                        make_expression::<T>(&parsed_tokens[i + n_uops + 1..], &parsed_vars, uop)?;
+                    let (expr, i_forward) = make_expression::<T>(
+                        &parsed_tokens[i + n_uops + 1..],
+                        &parsed_vars,
+                        UnaryOpWithReprs {
+                            reprs: vec_of_uop_reprs,
+                            op: uop,
+                        },
+                    )?;
                     Ok((DeepNode::Expr(expr), i_forward + n_uops + 1))
                 }
             },
             ParsedToken::Var(name) => {
                 let expr = DeepEx::new(
                     vec![DeepNode::Var(find_var_index(&name))],
-                    BinOpVec::new(),
-                    uop,
+                    BinOpsWithReprs {
+                        reprs: Vec::new(),
+                        ops: BinOpVec::new(),
+                    },
+                    UnaryOpWithReprs {
+                        reprs: vec_of_uop_reprs,
+                        op: uop,
+                    },
                 )?;
                 Ok((DeepNode::Expr(expr), n_uops + 1))
             }
@@ -248,6 +269,7 @@ where
     };
 
     let mut bin_ops = BinOpVec::new();
+    let mut reprs_bin_ops: Vec<&str> = Vec::new();
     let mut nodes = Vec::<DeepNode<T>>::new();
 
     // The main loop checks one token after the next whereby sub-expressions are
@@ -256,16 +278,17 @@ where
     let mut idx_tkn: usize = 0;
     while idx_tkn < parsed_tokens.len() {
         match &parsed_tokens[idx_tkn] {
-            ParsedToken::Op(b) => match b.unary_op {
+            ParsedToken::Op(op) => match op.unary_op {
                 None => {
-                    bin_ops.push(unpack_binop(b.bin_op));
+                    bin_ops.push(unpack_binop(op.bin_op));
+                    reprs_bin_ops.push(op.repr);
                     idx_tkn += 1;
                 }
                 Some(uo) => {
                     // might the operator be unary?
                     if idx_tkn == 0 {
                         // if the first element is an operator it must be unary
-                        let (node, idx_forward) = process_unary(idx_tkn, uo)?;
+                        let (node, idx_forward) = process_unary(idx_tkn, uo, op.repr)?;
                         nodes.push(node);
                         idx_tkn += idx_forward;
                     } else {
@@ -273,7 +296,8 @@ where
                         match &parsed_tokens[idx_tkn - 1] {
                             ParsedToken::Num(_) | ParsedToken::Var(_) => {
                                 // number or variable as predecessor means binary operator
-                                bin_ops.push(unpack_binop(b.bin_op));
+                                bin_ops.push(unpack_binop(op.bin_op));
+                                reprs_bin_ops.push(op.repr);
                                 idx_tkn += 1;
                             }
                             ParsedToken::Paren(p) => match p {
@@ -282,12 +306,13 @@ where
                                     panic!("{}", msg);
                                 }
                                 Paren::Close => {
-                                    bin_ops.push(unpack_binop(b.bin_op));
+                                    bin_ops.push(unpack_binop(op.bin_op));
+                                    reprs_bin_ops.push(op.repr);
                                     idx_tkn += 1;
                                 }
                             },
                             ParsedToken::Op(_) => {
-                                let (node, idx_forward) = process_unary(idx_tkn, uo)?;
+                                let (node, idx_forward) = process_unary(idx_tkn, uo, op.repr)?;
                                 nodes.push(node);
                                 idx_tkn += idx_forward;
                             }
@@ -309,7 +334,10 @@ where
                     let (expr, i_forward) = make_expression::<T>(
                         &parsed_tokens[idx_tkn..],
                         &parsed_vars,
-                        UnaryOp::new(),
+                        UnaryOpWithReprs {
+                            reprs: Vec::new(),
+                            op: UnaryOp::new(),
+                        },
                     )?;
                     nodes.push(DeepNode::Expr(expr));
                     idx_tkn += i_forward;
@@ -321,7 +349,17 @@ where
             },
         }
     }
-    Ok((DeepEx::new(nodes, bin_ops, unary_ops)?, idx_tkn))
+    Ok((
+        DeepEx::new(
+            nodes,
+            BinOpsWithReprs {
+                reprs: reprs_bin_ops,
+                ops: bin_ops,
+            },
+            unary_ops,
+        )?,
+        idx_tkn,
+    ))
 }
 
 /// Tries to give useful error messages for invalid constellations of the parsed tokens
@@ -343,11 +381,21 @@ where
             msg: "cannot parse empty string".to_string(),
         });
     };
-    let num_pred_succ = |idx: usize, forbidden: Paren| match &parsed_tokens[idx] {
+
+    enum NeighborType {
+        Predecessor,
+        Successor,
+    }
+
+    let num_pred_succ = |idx: usize, neighbor_type: NeighborType| match &parsed_tokens[idx] {
         ParsedToken::Num(_) => Err(ExParseError {
             msg: "a number/variable cannot be next to a number/variable".to_string(),
         }),
         ParsedToken::Paren(p) => {
+            let forbidden = match neighbor_type {
+                NeighborType::Predecessor => Paren::Close,
+                NeighborType::Successor => Paren::Open,
+            };
             if p == &forbidden {
                 Err(ExParseError {
                     msg: "wlog a number/variable cannot be on the right of a closing parenthesis"
@@ -357,6 +405,21 @@ where
                 Ok(0)
             }
         }
+        ParsedToken::Op(op) => match neighbor_type {
+            NeighborType::Predecessor => Ok(0),
+            NeighborType::Successor => {
+                if let Some(_) = op.bin_op {
+                    Ok(0)
+                } else if let Some(_) = op.unary_op {
+                    Err(ExParseError {
+                        msg: "a number/variable cannot be on the left of a unary operator"
+                            .to_string(),
+                    })
+                } else {
+                    Ok(0)
+                }
+            }
+        },
         _ => Ok(0),
     };
     let binop_pred_succ = |idx: usize| match parsed_tokens[idx] {
@@ -391,10 +454,10 @@ where
             match expr_elt {
                 ParsedToken::Num(_) | ParsedToken::Var(_) => {
                     if i < parsed_tokens.len() - 1 {
-                        num_pred_succ(i + 1, Paren::Open)?;
+                        num_pred_succ(i + 1, NeighborType::Successor)?;
                     }
                     if i > 0 {
-                        num_pred_succ(i - 1, Paren::Close)?;
+                        num_pred_succ(i - 1, NeighborType::Predecessor)?;
                     }
                     Ok(0)
                 }
@@ -439,9 +502,9 @@ where
     }
 }
 
-fn parsed_tokens_to_flatex<T: Copy + FromStr + Debug>(
-    parsed_tokens: &SmallVec<[ParsedToken<T>; 2 * N_NODES_ON_STACK]>,
-) -> Result<FlatEx<T>, ExParseError> {
+fn parsed_tokens_to_deepex<'a, T: Copy + FromStr + Debug>(
+    parsed_tokens: &SmallVec<[ParsedToken<'a, T>; 2 * N_NODES_ON_STACK]>,
+) -> Result<DeepEx<'a, T>, ExParseError> {
     let mut found_vars = SmallVec::<[&str; 16]>::new();
     let parsed_vars = parsed_tokens
         .iter()
@@ -460,8 +523,15 @@ fn parsed_tokens_to_flatex<T: Copy + FromStr + Debug>(
 
     check_preconditions(&parsed_tokens[..])?;
 
-    let (expr, _) = make_expression(&parsed_tokens[0..], &parsed_vars, UnaryOp::new())?;
-    Ok(expr.flatten())
+    let (expr, _) = make_expression(
+        &parsed_tokens[0..],
+        &parsed_vars,
+        UnaryOpWithReprs {
+            reprs: vec![],
+            op: UnaryOp::new(),
+        },
+    )?;
+    Ok(expr)
 }
 
 /// Parses a string and a vector of operators into an expression that can be evaluated.
@@ -470,13 +540,13 @@ fn parsed_tokens_to_flatex<T: Copy + FromStr + Debug>(
 ///
 /// An error is returned in case [`parse_with_number_pattern`](parse_with_number_pattern)
 /// returns one.
-pub fn parse<'a, T>(text: &str, ops: &[Operator<'a, T>]) -> Result<FlatEx<T>, ExParseError>
+pub fn parse<'a, T>(text: &str, ops: &[Operator<'a, T>]) -> Result<FlatEx<'a, T>, ExParseError>
 where
     <T as std::str::FromStr>::Err: Debug,
     T: Copy + FromStr + Debug,
 {
-    let parsed_tokens = parsed_tokens(text, ops, is_numeric_text)?;
-    parsed_tokens_to_flatex(&parsed_tokens)
+    let parsed_tokens = tokenize_and_analyze(text, ops, is_numeric_text)?;
+    Ok(flatten(parsed_tokens_to_deepex(&parsed_tokens)?))
 }
 
 /// Parses a string and a vector of operators and a regex pattern that defines the looks
@@ -514,7 +584,7 @@ pub fn parse_with_number_pattern<'a, 'b, T>(
     text: &'b str,
     ops: &[Operator<'a, T>],
     number_regex_pattern: &str,
-) -> Result<FlatEx<T>, ExParseError>
+) -> Result<FlatEx<'a, T>, ExParseError>
 where
     <T as std::str::FromStr>::Err: Debug,
     T: Copy + FromStr + Debug,
@@ -529,8 +599,8 @@ where
         }
     };
     let is_numeric = |text: &'b str| is_numeric_regex(&re_number, &text);
-    let parsed_tokens = parsed_tokens(text, ops, is_numeric)?;
-    parsed_tokens_to_flatex(&parsed_tokens)
+    let parsed_tokens = tokenize_and_analyze(text, ops, is_numeric)?;
+    Ok(flatten(parsed_tokens_to_deepex(&parsed_tokens)?))
 }
 
 /// Parses a string into an expression that can be evaluated using default operators.
@@ -551,15 +621,61 @@ where
 #[cfg(test)]
 mod tests {
     use crate::{
-        parse::{check_preconditions, is_numeric_text, make_default_operators, parsed_tokens},
-        ExParseError,
+        parse::{
+            check_preconditions, is_numeric_text, make_default_operators, parsed_tokens_to_deepex,
+            tokenize_and_analyze,
+        },
+        parse_with_default_ops, ExParseError,
     };
 
+
+        
+    #[test]
+    fn test_display() {
+        let mut flatex= parse_with_default_ops::<f64>("sin(var)/5").unwrap();
+        assert_eq!(format!("{}", flatex), "sin({x0})/5.0");
+        flatex.clear_deepex();
+        assert_eq!(format!("{}", flatex), "[FlatEx display information not available]");            
+    }
+
+    #[test]
+    fn test_unparse() {
+        fn test(text: &str, text_ref: &str) {
+            let ops = make_default_operators::<f64>();
+            let parsed = tokenize_and_analyze(text, &ops, is_numeric_text).unwrap();
+            let deepex = parsed_tokens_to_deepex(&parsed).unwrap();
+
+            assert_eq!(deepex.unparse(), text_ref);
+            let mut flatex_reparsed = parse_with_default_ops::<f64>(text_ref).unwrap();
+            assert_eq!(flatex_reparsed.unparse().unwrap(), text_ref);
+            flatex_reparsed.clear_deepex();
+            assert!(flatex_reparsed.unparse().is_err());
+        }
+        let text = "5+x";
+        let text_ref = "5.0+{x0}";
+        test(text, text_ref);
+        let text = "sin(5+var)^(1/{y})+{var}";
+        let text_ref = "sin(5.0+{x0})^(1.0/{x1})+{x0}";
+        test(text, text_ref);
+        let text = "-(5+var)^(1/{y})+{var}";
+        let text_ref = "-(5.0+{x0})^(1.0/{x1})+{x0}";
+        test(text, text_ref);
+        let text = "cos(sin(-(5+var)^(1/{y})))+{var}";
+        let text_ref = "cos(sin(-(5.0+{x0})^(1.0/{x1})))+{x0}";
+        test(text, text_ref);
+        let text = "cos(sin(-5+var^(1/{y})))-{var}";
+        let text_ref = "cos(sin(-5.0+{x0}^(1.0/{x1})))-{x0}";
+        test(text, text_ref);
+        let text = "cos(sin(-z+var*(1/{y})))+{var}";
+        let text_ref = "cos(sin(-({x0})+{x1}*(1.0/{x2})))+{x1}";
+        test(text, text_ref);
+    }
+
     #[test]
     fn test_apply_regexes() {
         let text = r"5\6";
         let ops = make_default_operators::<f32>();
-        let elts = parsed_tokens(text, &ops, is_numeric_text);
+        let elts = tokenize_and_analyze(text, &ops, is_numeric_text);
         assert!(elts.is_err());
     }
 
@@ -576,7 +692,7 @@ mod tests {
                 }
             }
             let ops = make_default_operators::<f32>();
-            let elts = parsed_tokens(text, &ops, is_numeric_text);
+            let elts = tokenize_and_analyze(text, &ops, is_numeric_text);
             match elts {
                 Ok(elts_unwr) => {
                     let err = check_preconditions(&elts_unwr[..]);
@@ -585,7 +701,6 @@ mod tests {
                 Err(_) => check_err_msg(elts, msg_part),
             }
         }
-
         test("", "empty string");
         test("++", "the last element cannot be an operator");
         test(
@@ -605,5 +720,9 @@ mod tests {
             r"3. .4",
             r"a number/variable cannot be next to a number/variable",
         );
+        test(
+            r"2sin({x})",
+            r"number/variable cannot be on the left of a unary",
+        );
     }
 }
